(*
The window style names information of the Window Contents Explorer utility
by Alexei Hmelnov.
----------------------------------------------------------------------------
E-Mail: alex@monster.icc.ru
http://monster.icc.ru/~alex/
----------------------------------------------------------------------------

See the file "readme.txt" for more details.

------------------------------------------------------------------------
                             IMPORTANT NOTE:
This software is provided 'as-is', without any expressed or implied warranty.
In no event will the author be held liable for any damages arising from the
use of this software.
Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:
1. The origin of this software must not be misrepresented, you must not
   claim that you wrote the original software.
2. Altered source versions must be plainly marked as such, and must not
   be misrepresented as being the original software.
3. This notice may not be removed or altered from any source
   distribution.
*)

unit StlNames;

interface

uses
  SysUtils,Classes;

type
  PStyleNames = ^TStyleNames;
  TStyleNames = array[0..15] of PChar;

  PStyleRange=^TStyleRange;
  TStyleRange=record
    Next: PStyleRange;
    Base,Mask,SzUsed: integer;
    Names: PStyleNames;
  end ;

  PCtlStyleRec = ^TCtlStyleRec;
  TCtlStyleRec = record
    RefCnt: integer;
    Ranges: PStyleRange;
    Names: TStyleNames;
  end ;

procedure InitWinClassDescrs;
procedure DoneWinClassDescrs;

function GetWinClassDescr(Name: String): PCtlStyleRec;

implementation

const
  WinClasses: TStringList = Nil;

procedure FreeStyleNames(Cnt: integer; Names: PStyleNames);
begin
  if Names=Nil then
    Exit;
  while Cnt>0 do begin
    Dec(Cnt);
    StrDispose(Names^[Cnt]);
  end ;
end ;

procedure FreeCtlStyleRec(CSR: PCtlStyleRec);
var
  RP,RPNext: PStyleRange;
  NDX: integer;
begin
  if CSR=Nil then
    Exit;
  Dec(CSR^.RefCnt);
  if CSR^.RefCnt>0 then
    Exit;
  RPNext := CSR^.Ranges;
  while RPNext<>Nil do begin
    RP := RPNext;
    RPNext := RP^.Next;
    if RP^.Names<>Nil then begin
      FreeStyleNames(RP^.SzUsed,RP^.Names);
      FreeMem(RP^.Names,RP^.SzUsed*SizeOf(Pointer));
    end ;
    Dispose(RP);
  end ;
  FreeStyleNames(16,@CSR^.Names);
  Dispose(CSR);
end ;

function NewCtlStyleRec: PCtlStyleRec;
begin
  Result := AllocMem(SizeOf(TCtlStyleRec));
end ;

function NewRange(CSR: PCtlStyleRec; Base,Mask: integer): PStyleRange;
begin
  New(Result);
  Result^.Next := CSR^.Ranges;
  CSR^.Ranges := Result;
  Result^.Base := Base;
  Result^.Mask := Mask;
  Result^.SzUsed := Result^.Mask+1;
  Result^.Names := AllocMem(Result^.SzUsed*SizeOf(Pointer));
end ;

function IsPwr2(Al: LongInt): integer;
var
  Pwr: integer;
begin
  IsPwr2 := -1;
  if Al=0 then
    Exit;
  Pwr := 0;
  while Al and 1=0 do begin
    Al := Al shr 1;
    Inc(Pwr);
  end ;
  if Al=1 then
    IsPwr2 := Pwr
  else
    IsPwr2 := -1-Pwr
end ;

procedure LoadClassStyles(FName: String; Lst: TStringList);

var
  l: integer;
  ScanP,ScanP0,EP: PChar;

  procedure ScanError(Msg: String);
  var
    S: String;
  begin
    S := StrPas(ScanP0);
    Insert('|',S,ScanP-ScanP0+1);
    raise Exception.CreateFmt('%s at %d:'#13#10'"%s"',
      [Msg,l,S]);
  end ;

  procedure ScanErrorFmt(Msg: String; Args: array of Const);
  begin
    ScanError(Format(Msg,Args));
  end ;

  procedure SkipSpace;
  var
    CP: PChar;
  begin
    CP := ScanP;
    while (CP^<>#0)and((CP^=' ')or(CP^=#9)) do
      Inc(CP);
    if CP^=';' then
      CP := StrEnd(CP);
    ScanP := CP;
  end ;

  function NextLine: boolean;
  var
    ZP: PChar;
  begin
    repeat
      Result := EP<>Nil;
      if not Result then
        Exit;
      ScanP := EP;
      EP := StrScan(ScanP,#10);
      Inc(l);
      if EP<>Nil then begin
        ZP := EP;
        Inc(EP);
        if (ZP>ScanP)and((ZP-1)^=#13) then
          ZP := ZP-1;
        ZP^ := #0;
      end ;
      ScanP0 := ScanP;
      SkipSpace;
    until ScanP^<>#0;
  end ;

  procedure NextLineChk;
  begin
    if not NextLine then
      ScanError('unexpected end of file');
  end ;

  function ReadValue: LongInt;
  var
    AP: PChar;
    V: LongInt;
    ch : Char;
    D: integer;
    Radix: integer;
  begin
    V := 0;
    AP := ScanP;
    ch := AP^;
    radix := 10;
    if ch='0' then begin
      Inc(AP);
      ch := UpCase(AP^);
      Inc(AP);
      Case ch of
       'X': radix := 16;
       'O': radix := 8;
       'B': radix := 2;
      else
        radix := 8;
        Dec(AP);
      End ;
    end ;
    while true do begin
      ch := AP^;
      case ch of
        '0'..'9': D := Ord(ch)-Ord('0');
        'A'..'F': D := 10+Ord(ch)-Ord('A');
        'a'..'f': D := 10+Ord(ch)-Ord('a');
      else
        break;
      end ;
      if D>=radix then
        ScanErrorFmt('Digit in 0..%d expected',[radix]);
      V := V*radix+D;
      Inc(AP);
    end ;
    if AP=ScanP then
      ScanError('number expected');
    ScanP := AP;
    SkipSpace;
    Result := V;
  end ;

  function ReadName: String;
  var
    NP: PChar;
  begin
    NP := ScanP;
    if not(NP^ in ['A'..'Z','a'..'z','-','#','_']) then
      ScanError('identifier expected');
    Inc(ScanP);
    if NP^<>'-' then
     while (ScanP^ in ['A'..'Z','a'..'z','0'..'9','_']) do
       Inc(ScanP);
    SetString(Result,NP,ScanP-NP);
    SkipSpace;
  end ;

  procedure ChkChar(Ch: Char);
  begin
    if ScanP^<>ch then
      ScanErrorFmt('"%s" expected',[ch]);
    Inc(ScanP^);
    SkipSpace;
  end ;

var
  AllTxt: String;
  F: File;
  FS: LongInt;
  V: integer;
var
  Name: String;
  CNDX: integer;
  CSR: PCtlStyleRec;

  procedure ReadValueRange(RM: integer);
  var
    RBase,RMask: integer;
    RP: PStyleRange;
    Name: String;
    V: integer;
  begin
    RBase := IsPwr2(V);
    if RBase<0 then
      RBase := -RBase-1;
    RMask := V shr RBase;
    if RMask>255 then
      ScanErrorFmt('too large range : %d (>256)',[RMask+1]);
    RP := NewRange(CSR,RBase,RMask);
    while true do begin
      NextLineChk;
      if ScanP^=')' then
        Break;
      Name := ReadName;
      V := ReadValue;
      if (V and not RM)<>0 then
        ScanErrorFmt('value 0x%x out of range 0x%x',[V,RM]);
      RP^.Names^[V shr RBase] := StrNew(PChar(Name));
    end ;
  end ;

  procedure ReadFlags;
  var
    V,Pwr,UsedMask: integer;
    Name: String;
  begin
    UsedMask := $FFFF0000;
    while true do begin
      NextLineChk;
      if ScanP^=')' then
        Break;
      Name := ReadName;
      V := ReadValue;
      if UsedMask and V<>0 then
        ScanErrorFmt('value 0x%x intersects with previous values or out of range',[V]);
      UsedMask := UsedMask or V;
      if ScanP^<>'(' then begin {Read single flag}
        Pwr := IsPwr2(V);
        if Pwr<0 then
          ScanErrorFmt('power of 2 expected: 0x%x',[V]);
        CSR^.Names[Pwr] := StrNew(PChar(Name));
       end
      else
        ReadValueRange(V);
    end ;
  end ;

begin
  Assign(F,FName);
  Reset(F,1);
  try
    FS := FileSize(F);
    SetLength(AllTxt,FS);
    BlockRead(F,PChar(AllTxt)^,FS);
  finally
    Close(F);
  end ;
  EP := PChar(AllTxt);
  l := 0;
  while NextLine do begin
    CSR := Nil;
    repeat
      Name := ReadName;
      if WinClasses.Find(Name,CNDX) then
        ScanErrorFmt('class "%s" redefined',[Name]);
      if CSR=Nil then
        CSR := NewCtlStyleRec;
      Inc(CSR^.RefCnt);
      WinClasses.AddObject(Name,Pointer(CSR));
      if ScanP^<>',' then
        break;
      Inc(ScanP);
      SkipSpace;
    until false;
    ChkChar('(');
    ReadFlags;
  end ;
end ;

procedure InitWinClassDescrs;
begin
  if WinClasses<>Nil then
    Exit;
  WinClasses := TStringList.Create;
  WinClasses.Duplicates := dupError;
  WinClasses.Sorted := true;
  LoadClassStyles(ExtractFilePath(ParamStr(0))+'ClasStl.cfg',WinClasses);
end ;

procedure DoneWinClassDescrs;
var
  i: integer;
begin
  if WinClasses=Nil then
    Exit;
  for i:=0 to WinClasses.Count-1 do
    FreeCtlStyleRec(Pointer(WinClasses.Objects[i]));
  WinClasses.Free;
  WinClasses := Nil;
end ;

function GetWinClassDescr(Name: String): PCtlStyleRec;
var
  NDX: integer;
begin
  Result := Nil;
  if WinClasses=Nil then
    Exit;
  if not WinClasses.Find(Name,NDX) then
    Exit;
  Result := Pointer(WinClasses.Objects[NDX]);
end ;


end.
